__version__ = "0.1.4"

__title__ = "joulescope_server"
__description__ = 'Joulescope™ server'
__url__ = 'https://joulescope.readthedocs.io'
__author__ = 'Jetperch LLC'
__author_email__ = 'joulescope-dev@jetperch.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2020-2022 Jetperch LLC'
